/*****************************************************************
        This module is the 2nd stage of the VeSPA processor.
        The following modules are instantiated
        1. opcode decoder
        2. sign extension module
        3. check condition codes
******************************************************************/
module stage2(IR2, PC2, clk, C, Z, N, V, R1, R2, Z5, clr, nop3, pause3, c5a, c6a, c7,  a1, a2, IR3, PC3, X3, Y3, MD3, cond_bit, COND2, RW3, HALT3);   
   parameter gate_delay = 3;   
   input [31:0]  IR2;      // Instruction Register from pipeline stage 1.
   input [31:0]  PC2;      // Program Counter from pipeline stage 1.
   input 	 clk;      // The clock.
   input 	 C;        // Status bit - Carry - from pipeline stage 3.
   input 	 Z;        // Status bit - Zero - from pipeline stage 3.
   input 	 N;        // Status bit - Negative - from pipeline stage 3.
   input 	 V;        // Status bit - oVerflow - from pipeline stage 3.
   input [31:0]  R1;       // A source operand from the register file.
   input [31:0]  R2;       // A source operand from the register file.
   input 	 clr;      // The clear signal used to reset all D-type flip-flops.
   input 	 nop3;     // If nop3 == 1'b1 then IR3 will be a NOP instruction.
   input 	 pause3;   // If pause3 == 1'b1 then IR3 is fed back into IR3.
   input 	 c5a;      // Used as part of the selector for the ALU operand 1. 
   input 	 c6a;      // Used as part of the selector for the ALU operand 2.
   input 	 c7;       // Selector for the md3 mux.
   input [31:0]  Z5;       // The result from pipeline stage 5.	            
   output [4:0]  a1;       // The address of source operand R1. 
   output [4:0]  a2;       // The address of source operand R2.
   output [31:0] IR3;      // Instruction Register for pipeline stage 3.
   output [31:0] PC3;      // Program Counter for pipeline stage 3.
   output [31:0] X3;       // A source operand for pipeline stage 3.
   output [31:0] Y3;       // A source operand for pipeline stage 3.
   output [31:0] MD3;      // The operand to store in memory for pipeline stage 3. 
   output 	 cond_bit; // The COND2 bit for for pipeline stage 3. This is in a flip-flop
   output 	 COND2;    // The cond_bit for the instruction in pipeline stage 2. This is not in a flip-flop.
   output 	 RW3;      // BWAT: The Register Write bit for pipeline stage 3.
   output 	 HALT3;    // BWAT: The halt bit for pipeline stage 3.

   //defining some internal wires
   wire 	 mp2;
   wire 	 imm2;
   wire 	 mp3;
   wire 	 mp4;
   wire [4:0] 	 op_a2mux;
   // connecting wires
   //wire[31:0] R1,R2, 
   wire [31:0] 	 PC3_mux_in;
   wire [31:0] 	 out2;
   wire [31:0] 	 X3_in;
   wire [31:0] 	 Y3_in;
   wire 	 alu2;
   wire [4:0] 	 a1,a2;
   wire 	 clr;   
   wire 	 tmp_cond;
   wire 	 COND2;	
   wire [31:0] 	 ir3_mux;
   wire [31:0] 	 nop3_out;
   wire [31:0] 	 md3_mux;
   
   //Instantiate OPCODE Decoder
   // delay of 6 inside module 
   dcd5_32 opdecode2(IR2[31:27], nop2, add2, sub2, or2, and2, not2, xor2, cmp2, br2, jmp2, ld2, ldi2, ldx2, st2, stx2, hlt2);
   
   // Instantiate Sign Extension Module
   // parameter is the individual gate delays for all gates in sext module
   sext_mod #(3)sext(IR2, alu2, jmp2, br2, st2, ld2, ldi2, stx2, ldx2,  out2, imm16, imm17, imm22, imm23);
   
   // Instantiate checkcc module
   testcc checkcc(IR2[26:23], C, Z, N, V, tmp_cond);
   
   // LOGIC FOR PAUSING THE X3 AND Y3 MUX's WHEN PAUSE 3 signal is active
   wire 	 pause3b;
   
   not #3 (pause3b, pause3);
   or #3 (mp2_c5a, mp2, c5a);
   or #3 (mp3_c6a, mp3, c6a);
   and #3 (x3sel0, pause3b, mp2_c5a);
   or #3 (x3sel1, pause3, c5a);
   and #3(y3sel0, pause3b, mp3_c6a);
   or #3 (y3sel1, pause3, c6a);
   
   
   // Instantiate Mux's
   // select, Output, Input0, Input1, in2,in3
   mux4 #(32,6) x3mux(x3sel0, x3sel1, X3_in, R1, PC2, X3, Z5);
   mux4 #(32,6) y3mux(y3sel0, y3sel1, Y3_in, R2, out2, Y3, Z5);
   mux2 #(5,6)  a2mux(mp4, op_a2mux, IR2[15:11], IR2[26:22]);
   mux4 #(32,6)  md3mux(c7, pause3, md3_mux, R2, Z5, MD3, MD3);
   
   //mux4 sel0, sel1, out, in0-3! selects what goes in IR3 
   mux4 #(32,6) ir3mux(nop3, pause3, ir3_mux, IR2, nop3_out, IR3, IR3);
   
   //mux for PC2 input
   mux2 #(32,6) pc3_mux(pause3, PC3_mux_in, PC2, PC3);
   
   //Instantiate condition register
   // parameters width, propogation delay
   d_reg #(1,8) condreg(clk, cond_bit, COND2, clr);
   
   // dreg for nop3
   d_reg #(32,8) nop3reg(clk, nop3_out, 0, clr);  // A NOP instruction.
   
   // instantiate 32 bit register
   // clock output input
   d_reg #(32,8) PC3_reg (clk, PC3, PC3_mux_in, clr);
   d_reg #(32,8) X3_reg  (clk, X3, X3_in, clr);
   d_reg #(32,8) Y3_reg  (clk, Y3, Y3_in, clr);
   d_reg #(32,8) IR3_reg (clk, IR3, ir3_mux, clr);
   d_reg #(32,8) MD3_reg (clk, MD3, md3_mux, clr); 
         
   // GENERATE PROPER CONDITION CODE
   and #(gate_delay) (COND2, br2, tmp_cond);
      
   // COMBINATIONAL CIRCUITS AND OTHER ASSIGNMENTS	
   
   // defining a1 and a2 signals
   assign a1 = IR2[21:17];
   assign a2 = op_a2mux;
   
   // creating signals for mux's
   assign mp2 = br2;
   or #(gate_delay) imm_or(imm2, imm16, imm17, imm22, imm23);	
   assign mp3 = imm2;
   or #(gate_delay) st_or_stx(mp4, st2, stx2);
   
   // creating alu2 signal
   or #(gate_delay) gen_alu2 (alu2, add2, sub2, or2, and2, not2, xor2, cmp2);

   and #(gate_delay) gen_jmpl(jmpl, jmp2, IR2[16]);     // BWAT: JUMPL writes to a register
   or #(gate_delay) gen_write_or(RW2, add2, sub2, and2, 
				 or2, not2, xor2, jmpl, 
				 ld2, ldi2, ldx2);      // BWAT: So do these instructions
   d_reg #(1,8) rw3(clk, RW3, RW2, clr);                // BWAT: Pass on the Register Write bit to pipeline stage 3. 
   d_reg #(1,8) halt3(clk, HALT3, hlt2, clr);           // BWAT: Pass on the halt bit to pipeline stage 3.
endmodule // stage2













